import itertools
import logging
import random
from pathlib import Path
from typing import List, Tuple

import numpy as np
import tifffile as tf
from attrs import define, field, validators

logger = logging.getLogger(__name__)


def path_exists(instance, attribute, value):
    if not value.exists():
        raise FileNotFoundError(f"Invalid path ({value})")


def has_projections(instance, attribute, value):
    projections = value / "projections"
    if not projections.is_dir():
        raise FileNotFoundError(f"Invalid path for Projections: ({value}).")


def is_tif(instance, attribute, value):
    if value.suffix != ".tif":
        raise ValueError(f"Not a tif file ({value})")


@define
class Field:
    """
    Represents a Field

    It is responsible for returning the field's name and to load the corresponding image.

    Attributes:
        path: path to the image file
    """

    path: Path = field(validator=[validators.instance_of(Path), path_exists, is_tif])

    @property
    def name(self) -> str:
        """
        Extracts the field name as a string without the extension.
        """
        return self.path.stem

    @property
    def data(self) -> np.ndarray:
        """
        Loads data from path as a numpy array using `tifffile.imread`.
        """
        return tf.imread(str(self.path))


@define
class Dataset:
    """
    Represents a Dataset folder.

    It iterates over the fields and provides a setup method to create the necessary subdirectories.

    Attributes:
       path: Path to the dataset folder.
    """
    path: Path = field(validator=[
        validators.instance_of(Path),
        path_exists,
        has_projections])

    @property
    def logs(self):
        return self.path / "logs"

    @property
    def projections(self):
        return self.path / "projections"

    @property
    def predictions(self):
        return self.path / "predictions"

    @property
    def annotations(self):
        return self.path / "annotations"

    @property
    def visualisation(self):
        return self.path / "visualisation"

    @property
    def statistics(self):
        return self.path / "statistics"

    @property
    def nuclei(self):
        return self.predictions / "nuclei"

    @property
    def centrioles(self):
        return self.predictions / "centrioles"

    @property
    def cilia(self):
        return self.predictions / "cilia"

    @property
    def assignment(self):
        return self.predictions / "assignment"

    def setup(self) -> None:
        """
        Creates the subdirectories inside the Dataset path if not existing.
        The directories visualisation, statistics, predictions,
        nuclei, centrioles, cilia and assignment are created.
        """

        self.logs.mkdir(exist_ok=True)
        self.visualisation.mkdir(exist_ok=True)
        self.statistics.mkdir(exist_ok=True)
        self.predictions.mkdir(exist_ok=True)
        self.annotations.mkdir(exist_ok=True)
        self.nuclei.mkdir(exist_ok=True)
        self.centrioles.mkdir(exist_ok=True)
        self.cilia.mkdir(exist_ok=True)
        self.assignment.mkdir(exist_ok=True)

    @property
    def fields(self) -> List[Field]:
        """
        Collects all Fields found in `projections` into a list and raises a ValueError if no TIF file.
        """

        result = []
        for path in self.projections.iterdir():
            if (path.suffix == '.tif') and not path.name.startswith("."):
                result.append(Field(path))

        if len(result) == 0:
            raise ValueError(f"No field found in {self.projections}")

        return result

    def split_pairs(self, channels: Tuple[int], p=0.9, seed=1993) -> tuple[list[tuple[Field, int]], list[tuple[Field, int]]]:
        """
        Split a list of pairs (field, channel).

        Args:
            channels: the channels to choose from
            p: the train proportion, default to .9
            seed: the seed to reproduce the splitting
        Returns:
            train_split, test_split
        """
        random.seed(seed)

        size = len(self.fields)
        split_idx = int(p * size)
        _channels = random.choices(channels, k=size)

        pairs = [(field, int(channel)) for field, channel in zip(self.fields, _channels)]

        shuffled = random.sample(pairs, k=size)
        split_train = shuffled[:split_idx]
        split_test = shuffled[split_idx:]

        return split_train, split_test

    def choose_channel(self, fields: list[Field], channels: list[int]) -> list[tuple[Field, int]]:
        """Pick a channel for each field."""
        return [(field, int(channel)) for field, channel in itertools.product(fields, channels)]



import argparse
import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Tuple, List

from numpy.random import seed
from spotipy.model import SpotNet, Config
from tensorflow.random import set_seed

from cenfind.core.data import Dataset, Field
from cenfind.core.loading import fetch_all_fields
from cenfind.training.config import config_multiscale, transforms

seed(1)
set_seed(2)

logger = logging.getLogger(__name__)


def read_config(path):
    """
    Read config json file.
    :param path:
    :return:
    """
    with open(path, "r") as config:
        config_dict = json.load(config)
    return Config(**config_dict)


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "train", help="Train a Spotnet model using the datasets"
    )
    parser.add_argument("datasets", type=Path, nargs="+", help="Path to the dataset")
    parser.add_argument("channels", type=int, nargs="+", help="the channel numbers to train on, e.g., 1 2 3")
    parser.add_argument(
        "--model_path", type=Path, required=True, default=Path(".").resolve(), help="Path to the model fit"
    )
    parser.add_argument("--epochs", type=int, required=True, default=1, help="Number of epochs")

    return parser


def write_split(split: List[Tuple[Field, int]], dst: Path) -> None:
    with open(dst, "w") as f:
        for field, channel in split:
            line = f"{field.name},{channel}\n"
            f.write(line)


def run(args):
    dataset = Dataset(args.dataset)

    split_train, split_test = dataset.split_pairs(channels=args.channels)
    if not (dataset.path / "train.txt").exists():
        write_split(split_train, dataset.path / "train.txt")
    if not (dataset.path / "test.txt").exists():
        write_split(split_test, dataset.path / "test.txt")

    all_train_x, all_train_y, all_test_x, all_test_y = fetch_all_fields(dataset, transforms=transforms)
    logger.debug("Loading %s" % (len(all_train_x)))

    if not args.model_path.exists():
        logger.info("Creating model folder (%s)" % (str(args.model_path)))
        args.model_path.mkdir()

    time_stamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    model_multiscale = SpotNet(
        config_multiscale, name=time_stamp, basedir=args.model_path
    )

    model_multiscale.train(
        all_train_x,
        all_train_y,
        validation_data=(all_test_x, all_test_y),
        epochs=args.epochs,
    )


if __name__ == "__main__":
    args = argparse.Namespace(dataset=Path("../../../data/RPE1wt_CEP63+CETN2+PCNT_1/"),
                              channels=(1,2,3),
                              model_path=Path("../../../models/"),
                              epochs=100)
    run(args)



import argparse
import logging

import pandas as pd
from pathlib import Path
import seaborn as sns
from matplotlib import pyplot as plt

from cenfind.core.data import Dataset
logging.basicConfig(level=logging.ERROR)

def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "analyse",
        help="Analyse predictions",
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset folder")

    return parser

def run(args):

    sns.set_theme()
    dataset = Dataset(args.dataset)
    # Concatenate the predictions and keep identity of field of view
    files = Path(dataset.predictions / "centrioles").glob('*.tsv')
    dfs = list()
    for f in files:
        data = pd.read_csv(f, sep='\t')
        data['file'] = f.stem
        dfs.append(data)

    df = pd.concat(dfs, ignore_index=True)

    # Plot the intensities
    g = sns.displot(df, x='intensity', hue='channel')
    g.savefig(dataset.statistics / "vis_intensities.png")

    # Plot the scores
    data = pd.read_csv(dataset.statistics / "statistics.tsv", sep='\t')
    summary = data.groupby('channel').sum(numeric_only=True)
    summary = pd.wide_to_long(summary.reset_index(), stubnames='', i='channel', j='number', suffix='(\d)|\+')
    summary.columns = ['frequency']
    summary = summary.reset_index()
    g = sns.catplot(summary, x='number', y='frequency', hue='channel', kind='bar')
    g.savefig(dataset.visualisation / 'vis_scores.png')

    # Plot SA/I of nuclei
    files = Path(dataset.predictions / 'nuclei').glob('*.json')
    dfs = list()
    for f in files:
        data = pd.read_json(f)
        data = pd.DataFrame(data["nuclei"].values.tolist()).drop("contour", axis=1)
        data['file'] = f.stem
        dfs.append(data)
    df = pd.concat(dfs, ignore_index=True)
    df = df.loc[df['is_nucleus_full']]

    f, ax = plt.subplots(figsize=(6.5, 6.5))
    sns.scatterplot(data=df, x='surface_area', y='intensity', ax=ax)
    sns.rugplot(data=df, x='surface_area', y='intensity', ax=ax)
    f.savefig(dataset.visualisation / 'vis_surface_area_intensity.png')

    f, ax = plt.subplots(figsize=(6.5, 6.5))
    sns.histplot(data=df, x='intensity', ax=ax, kde=True)
    f.savefig(dataset.visualisation / 'vis_intensity.png')


if __name__ == '__main__':
    args = argparse.Namespace(dataset=Path("../../../data/dataset_test"))
    run(args)



import contextlib
import functools
import logging
import os
from pathlib import Path
from typing import List

import cv2
import numpy as np
import tensorflow as tf
from csbdeep.utils import normalize
from skimage import measure
from skimage.exposure import rescale_intensity
from skimage.feature import hessian_matrix, hessian_matrix_eigvals
from skimage.filters.thresholding import threshold_otsu
from spotipy.model import SpotNet
from spotipy.utils import normalize_fast2d
from stardist.models import StarDist2D

from cenfind.core.data import Field
from cenfind.core.structures import Centriole, Nucleus
from cenfind.core.visualisation import draw_foci, resize_image

np.random.seed(1)
tf.random.set_seed(2)
tf.get_logger().setLevel(logging.ERROR)

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

logger = logging.getLogger(__name__)


def extract_foci(field: Field, channel: int, foci_model_file: Path,
                 prob_threshold=0.5, min_distance=2, ) -> List[Centriole]:
    """
    Detects centrioles in Field as (row, col) position.

    :param field: Field of view to search for centrioles.
    :param foci_model_file: SpotNet trained model file.
    :param channel: Channel to use in the field.
    :param prob_threshold: Probability threshold used for the cutoff (default: 0.5).
    :param min_distance: Minimal distance between two centrioles (default: 2 pixels).
    :return: List of centriole objects.
    """
    logger.info("Processing %s / %d" % (field.name, channel))
    if field.data.ndim != 3:
        raise ValueError("Bad data shape: %s; Ensure that the image is CXY" % field.data.shape)
    data = field.data[channel, ...]

    @functools.lru_cache(maxsize=None)
    def get_model(model):
        path = Path(model)
        if not path.is_dir():
            raise (FileNotFoundError(f"{path} is not a directory"))

        return SpotNet(None, name=path.name, basedir=str(path.parent))

    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
        data = normalize_fast2d(data)
        model = get_model(foci_model_file)
        _, points_preds = model.predict(
            data, prob_thresh=prob_threshold, min_distance=min_distance, verbose=False
        )

    foci = []
    for f_id, (r, c) in enumerate(points_preds.tolist()):
        foci.append(Centriole(field=field, channel=channel, centre=(r, c), index=f_id, label="Centriole"))

    centrosomes_mask = np.zeros(data.shape, dtype="uint8")
    centrosomes_mask = draw_foci(centrosomes_mask, foci, radius=min_distance * 2)

    centrosomes_map = measure.label(centrosomes_mask)
    centrosomes_centroids = measure.regionprops(centrosomes_map)

    for f in foci:
        foci_index = centrosomes_map[f.centre]
        centrosome_centroid = centrosomes_centroids[foci_index - 1].centroid
        f.parent = Centriole(field=field, channel=channel, centre=centrosome_centroid, label="Centrosome")

    if len(foci) == 0:
        logger.warning("No centrioles (channel: %s) has been detected in %s" % (channel, field.name))

    logger.info("(%s), channel %s: foci: %s" % (field.name, channel, len(foci)))
    return foci


def extract_nuclei(field: Field, channel: int, model: StarDist2D = None) -> List[Nucleus]:
    """
    Extracts the nuclei from the field.

    :param field: Field of view to search for nuclei.
    :param channel: Channel to use in the field.
    :param model: Model instance of StarDist.

    :return: List of Nuclei.

    """
    if model is None:
        from stardist.models import StarDist2D
        with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
            model = StarDist2D.from_pretrained("2D_versatile_fluo")

    if field.data.ndim == 2:
        data = field.data
    elif field.data.ndim == 3:
        data = field.data[channel, ...]
    else:
        raise ValueError("Bad data shape: %s; Ensure that the image is CXY" % field.data.shape)

    data_resized = resize_image(data)
    with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
        labels, _ = model.predict_instances(normalize(data_resized))
    labels = cv2.resize(
        labels, dsize=data.shape, fx=1, fy=1, interpolation=cv2.INTER_NEAREST
    )

    if len(labels) == 0:
        logger.warning("No nucleus has been detected in %s" % field.name)
        return []
    labels_id = np.unique(labels)

    nuclei = []
    for nucleus_index, nucleus_label in enumerate(labels_id):
        if nucleus_label == 0:
            continue
        sub_mask = np.zeros_like(labels, dtype="uint8")
        sub_mask[labels == nucleus_label] = 1
        contour, _ = cv2.findContours(
            sub_mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE
        )
        nucleus = Nucleus(field=field, channel=channel, contour=contour[0], label="Nucleus", index=nucleus_index - 1)
        nuclei.append(nucleus)

    logger.info("Nuclei extraction (channel %s) in %s: %s" % (channel, field.name, len(nuclei)))

    return nuclei


def extract_cilia(field: Field, channel, sigma=5.0, eccentricity=.9, area=200) -> List[Centriole]:
    """
    Extracts the cilia using the Hessian.

    :param field: Field of view to search for nuclei.
    :param channel: Channel to use in the field.
    :param sigma: Scale for the blob detection.
    :param eccentricity: Sphericity of the blobs to be detected.
    :param area: Filter area to use.
    :return: List of cilia Point objects.
    """
    data = field.data[channel, ...]
    resc = rescale_intensity(data, out_range="uint8")

    h_elems = hessian_matrix(resc, sigma=sigma, order="rc")
    _, minima_ridges = hessian_matrix_eigvals(h_elems)
    threshold = threshold_otsu(minima_ridges)

    mask = minima_ridges < threshold
    labels = measure.label(mask)
    props = measure.regionprops(labels, mask)

    result = []
    for prop in props:
        if prop.eccentricity > eccentricity and prop.area > area:
            r, c = prop.centroid
            result.append(Centriole(field=field, channel=channel, centre=(int(r), int(c)), label="Cilium"))

    logger.info("Cilium extraction (channel %s) in %s: %s" % (channel, field.name, len(result)))

    return result



from typing import Tuple
import logging

import cv2
import numpy as np
from attrs import define

from cenfind.core.data import Field

logger = logging.getLogger(__name__)


@define
class Centriole:
    """
    Represents a centriole as a 2d point associated with a Field and a channel index.

    The parent attributes is used to ensure that neighbouring centrioles are assigned to the
    same nucleus.

    Attributes:
        field: Field used for centriole detection.
        channel: Channel used for centriole detection.
        centre: Coordinates (row major) of the detected centriole.
        index: Index of centriole.
        label: String label
        parent: Parent centriole used for tying neighbouring centrioles into a centrosome.
    """
    field: Field
    channel: int
    centre: Tuple[int, int]
    index: int = 0
    label: str = ""
    parent: "Centriole" = None

    @property
    def centre_xy(self) -> Tuple[int, int]:
        """
        Represent a row-major point as an XY point.
        Returns:

        """
        y, x = self.centre

        return x, y

    def intensity(self, image: np.ndarray, k: int = 0, channel: int = None) -> int:
        """
        Computes the intensity of the focus within the k-pixel neighbourhood on the image provided.
        Args:
            image: The image for signal extraction.
            k: Pixel (radius) neighbourhood for signal extraction
            channel: Channel to use for signal extraction.

        Returns:

        """
        r, c = self.centre
        max_r, max_c = image.shape[-2:]

        r_start = np.clip(r - k, 0, max_r)
        r_stop = np.clip(r + k, 0, max_r)
        c_start = np.clip(c - k, 0, max_c)
        c_stop = np.clip(c + k, 0, max_c)

        if image.ndim < 3:
            return int(np.sum(image[r_start:r_stop, c_start:c_stop]))

        if channel is None:
            raise ValueError("Channel must be supplied when image has 3 dimensions")

        return int(np.sum(image[channel, r_start:r_stop, c_start:c_stop]))

    def as_dict(self) -> dict:
        return {
            "channel": self.channel,
            "pos_r": self.centre[0],
            "pos_c": self.centre[1],
            "intensity": self.intensity(self.field.data[self.channel, ...], k=1)
        }


@define
class Nucleus:
    """
    Represents a Nucleus with a contour.

    Attributes:
        field: Field of view
        channel: Channel used for nucleus segmentation
        contour: Coordinates of the Nucleus contour.
        index: Index of the nucleus.
        label: Label of the Nucleus
    """
    field: Field
    channel: int
    contour: np.ndarray
    index: int = 0
    label: str = ""

    @property
    def centre(self) -> Tuple[int, int]:
        """
        Centroid of the Nucleus as row-major.
        """
        moments = cv2.moments(self.contour)
        centre_x = int(moments["m10"] / (moments["m00"] + 1e-5))
        centre_y = int(moments["m01"] / (moments["m00"] + 1e-5))

        return int(centre_y), int(centre_x)

    @property
    def centre_xy(self) -> Tuple[int, int]:
        """
        XY representation of the centre of the contour
        Returns:

        """
        y, x = self.centre

        return int(x), int(y)

    @property
    def intensity(self) -> int:
        _data = self.field.data[self.channel, ...]
        mask = np.zeros_like(_data)
        cv2.drawContours(mask, [self.contour], 0, (1,), -1)
        mask = mask > 0
        masked = _data.copy()
        masked[~mask] = 0

        return int(np.sum(masked))

    @property
    def area(self) -> int:
        return int(cv2.contourArea(self.contour))

    @property
    def full_in_field(self) -> bool:
        """
        Checks if a contour is fully visible.

        Specifically, it checks whether its centre is within the 5-% margin of the image.

        """
        h, w = self.field.data.shape[-2:]
        fraction = 0.05
        pad_lower = int(fraction * h)
        pad_upper = h - pad_lower
        centroid = self.centre

        if all([pad_lower < c < pad_upper for c in centroid]):
            return True

        return False

    def as_dict(self) -> dict:
        """
        Converts a Contour into a dictionary for serialisation.

        Returns: Dictionary with keys: channel, pos_r, pos_c, intensity, surface area, is_nucleus_full, contour.

        """

        result = {
            "channel": self.channel,
            "pos_r": self.centre[0],
            "pos_c": self.centre[1],
            "intensity": self.intensity,
            "surface_area": self.area,
            "is_nucleus_full": self.full_in_field,
            "contour": self.contour.tolist()
        }

        return result



from datetime import datetime
from pathlib import Path

from spotipy.model import SpotNet

from cenfind.cli.train import fetch_all_fields
from cenfind.constants import datasets
from cenfind.core.data import Dataset
from cenfind.training.config import config_unet

PREFIX_REMOTE = Path("/data1/centrioles/canonical/")


def main():
    path_datasets = [PREFIX_REMOTE / ds for ds in datasets]
    dss = [Dataset(path) for path in path_datasets]

    all_train_x, all_train_y, all_test_x, all_test_y = fetch_all_fields(dss)

    time_stamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    model_unet = SpotNet(config_unet, name=time_stamp, basedir='models/dev/unet')
    model_unet.train(all_train_x, all_train_y, validation_data=(all_test_x, all_test_y), epochs=100)

    return 0


if __name__ == '__main__':
    main()



import contextlib
import os
from typing import Tuple

import albumentations as alb
import numpy as np
import tifffile as tf
from spotipy.utils import points_to_prob, normalize_fast2d

from cenfind.core.data import Dataset


def load_foci(path) -> np.ndarray:
    """
    Load annotation file from text file given channel
    loaded as row col, row major.
    ! the text format is x, y; origin at top left;
    :param channel:
    :return:
    """

    return np.loadtxt(str(path), dtype=int, delimiter=",")


def load_pairs(
        ds: Dataset, split: str, sigma: float = 1.5, transforms: alb.Compose = None
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Load two arrays, the images and the foci masks
    path: the path to the ds
    split: either train or test
    """

    channels = []
    masks = []

    with open(ds.path / f"{split}.txt", "r") as f:
        pairs = [l.strip("\n").split(",") for l in f.readlines()]

    for field, channel in pairs:
        data = tf.imread(ds.projections / f"{field}.tif")[int(channel), :, :]
        foci = load_foci(ds.annotations / "centrioles" / f"{field}_C{channel}.txt")

        with open(os.devnull, "w") as f, contextlib.redirect_stdout(f):
            image = normalize_fast2d(data, clip=True)

        if len(foci) == 0:
            mask = np.zeros(image.shape, dtype="uint16")
        else:
            mask = points_to_prob(foci, shape=image.shape, sigma=sigma)
        # from matplotlib import pyplot as plt
        # plt.imshow(image)
        # plt.show()
        # plt.imshow(mask)
        # plt.show()
        if transforms is not None:
            transformed = transforms(image=image, mask=mask)
            image = transformed["image"]
            mask = transformed["mask"]

        channels.append(image)
        masks.append(mask)

    return np.stack(channels), np.stack(masks)


def fetch_all_fields(dataset: Dataset, transforms: alb.Compose = None):
    all_train_x = []
    all_train_y = []

    all_test_x = []
    all_test_y = []

    train_x, train_y = load_pairs(dataset, split="train", transforms=transforms)
    test_x, test_y = load_pairs(dataset, split="test")
    all_train_x.append(train_x)
    all_train_y.append(train_y)
    all_test_x.append(test_x)
    all_test_y.append(test_y)

    all_train_x = np.concatenate(all_train_x, axis=0)
    all_train_y = np.concatenate(all_train_y, axis=0)

    all_test_x = np.concatenate(all_test_x, axis=0)
    all_test_y = np.concatenate(all_test_y, axis=0)

    return all_train_x, all_train_y, all_test_x, all_test_y



import logging
from typing import List

import cv2
import numpy as np
from skimage.draw import disk
from skimage.exposure import rescale_intensity

from cenfind.core.data import Field
from cenfind.core.structures import Centriole, Nucleus

logger = logging.getLogger(__name__)


def resize_image(image: np.ndarray, factor: int = 256) -> np.ndarray:
    """
    Resizes the image for nuclei segmentation by StarDist

    Args:
        image: a single channel image (H x W)
        factor: the target dimension

    Returns: Resized image

    """

    height, width = image.shape
    shrinkage_factor = int(height // factor)
    height_scaled = int(height // shrinkage_factor)
    width_scaled = int(width // shrinkage_factor)
    data_resized = cv2.resize(image,
                              dsize=(height_scaled, width_scaled),
                              fx=1, fy=1, interpolation=cv2.INTER_NEAREST,
                              )

    return data_resized


def draw_point(image: np.ndarray, point: Centriole,
               color: tuple[int, int, int] = (0, 255, 0), annotation: bool = True,
               marker_type: int = cv2.MARKER_SQUARE, marker_size: int = 8,
               ):
    """
    Draws a point object on an image.

    Args:
        image: Image to draw the point on.
        point: The point to draw
        color: The color to use (default: Green)
        annotation: whether to add text.
        marker_type: Marker type for point (default: square)
        marker_size: Marker size (default: 8 px)

    Returns: Annotated image.

    """
    r, c = point.centre
    offset_col = int(0.01 * image.shape[1])

    if annotation:
        cv2.putText(
            image,
            f"{point.label} {point.index}",
            org=(c + offset_col, r),
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=0.4,
            thickness=1,
            color=color,
        )

    return cv2.drawMarker(
        image, (c, r), color, markerType=marker_type, markerSize=marker_size
    )


def draw_foci(image: np.ndarray, foci: list[Centriole], radius=2) -> np.ndarray:
    """
    Draws foci as disks of given radius.

    Used to group centrioles into centrosomes.

    Args:
        image: Image to draw foci on
        foci: List of centrioles to draw.
        radius: Radius of the centriole as disk

    Returns: The mask image

    """

    mask = np.zeros(image.shape, dtype="uint8")
    for f in foci:
        rr, cc = disk(f.centre, radius, shape=image.shape[-2:])
        mask[rr, cc] = 250

    return mask


def draw_contour(image: np.ndarray, nucleus: Nucleus,
                 color: tuple[int, int, int] = (0, 255, 0), annotation: bool = True, thickness: int = 2):
    """
    Draws a contour from a Nucleus object on an image.

    Args:
        image: Image to draw on the contour
        nucleus: Contour to draw
        color: colour as tuple
        annotation: if True, add text to object
        thickness: Thickness of the contour line

    Returns: the masked image.

    """
    r, c = nucleus.centre
    cv2.drawContours(image, [nucleus.contour], -1, color, thickness=thickness)
    if annotation:
        cv2.putText(
            image,
            f"{nucleus.label}{nucleus.index}",
            org=(c, r),
            fontFace=cv2.FONT_HERSHEY_SIMPLEX,
            fontScale=0.8,
            thickness=2,
            color=color,
        )
        cv2.drawMarker(
            image, (c, r), color, markerType=cv2.MARKER_STAR, markerSize=10
        )
    return image


def _color_channel(image: np.ndarray, color: tuple[int, int, int], out_range: str) -> np.ndarray:
    """
    Creates a coloured version of a channel image

    Args:
        image: Image to convert
        color: Color to use (B, G, R)
        out_range: Depth, e.g., 'uint8'

    Returns:

    """
    """

    :param data: the data to use
    :param color: the color as a tuple (B, G, R)
    :param out_range:
    :return:
    """
    data = rescale_intensity(image, out_range=out_range)
    b = np.multiply(data, color[0], casting="unsafe")
    g = np.multiply(data, color[1], casting="unsafe")
    r = np.multiply(data, color[2], casting="unsafe")
    res = cv2.merge([b, g, r])

    return res


def create_vignette(field: Field, marker_index: int, nuclei_index: int) -> np.ndarray:
    """
    Normalises all markers and represent them as blue and highlight the channel in green.

    Args:
        field: Field to use
        marker_index: Channel for th marker
        nuclei_index: Channel

    Returns:

    """

    layer_nuclei = field.data[nuclei_index, ...]
    layer_marker = field.data[marker_index, ...]

    nuclei = _color_channel(layer_nuclei, (1, 0, 0), "uint8")
    marker = _color_channel(layer_marker, (0, 1, 0), "uint8")

    res = cv2.addWeighted(marker, 1, nuclei, 0.5, 0)
    res = cv2.putText(
        res,
        f"{field.name} {marker_index}",
        (100, 100),
        cv2.FONT_HERSHEY_SIMPLEX,
        0.8,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )
    return res


def visualisation(background: np.ndarray,
                  centrioles: List[Centriole],
                  nuclei: List[Nucleus],
                  assigned: np.ndarray = None,
                  ) -> np.ndarray:
    """
    Builds a visualisation image to assess the quality of the result.

    Args:
        background: The image to draw on.
        centrioles: Centrioles to be drawn.
        nuclei: Nuclei to be drawn.
        assigned: Arrows to draw between centrioles and nuclei.

    Returns: Annotated image

    """
    for centriole in centrioles:
        background = draw_point(background, centriole, annotation=False)

    for nucleus in nuclei:
        if nucleus.full_in_field:
            color = (0, 255, 0)
        else:
            color = (255, 0, 0)
        background = draw_contour(background, nucleus, color=color, annotation=False)

    if assigned is None:
        return background

    if not centrioles or not nuclei:
        return background

    it = np.nditer(assigned, flags=['multi_index'])
    for entry in it:
        if not entry:
            continue
        n, c = it.multi_index
        nucleus = nuclei[n]
        centriole = centrioles[c]
        cv2.arrowedLine(background, tuple(reversed(centriole.centre)), tuple(reversed(nucleus.centre)),
                        color=(0, 255, 0), thickness=2)

    return background



import sys
import argparse
from pathlib import Path
import logging
import numpy as np
import pandas as pd
import tifffile as tf

from cenfind.core.data import Dataset, Field
from cenfind.core.loading import load_foci
from cenfind.core.visualisation import visualisation, create_vignette
from cenfind.core.statistics import evaluate

ROOT_DIR = Path("../../../")

logger = logging.getLogger(__name__)

def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser(
        "evaluate", help="Evaluate the model on the test split of the dataset"
    )
    parser.add_argument("dataset", type=Path, help="Path to the dataset folder")
    parser.add_argument("model", type=Path, help="Path to the model")
    parser.add_argument(
        "--channel_nuclei",
        "-n",
        type=int,
        help="Channel index of nuclei",
    )
    parser.add_argument(
        "--tolerance",
        type=int,
        nargs="+",
        default=3,
        help="Distance in pixels below which two points are deemed matching",
    )
    parser.add_argument(
        "--threshold",
        type=float,
        default=.5,
        help="Confidence."
    )
    parser.add_argument(
        "--performances_file",
        type=Path,
        help="Path of the performance file, STDOUT if not specified",
    )

    return parser


def run(args):
    dataset = Dataset(args.dataset)
    dataset.setup()
    if not any((dataset.annotations / "centrioles").iterdir()):
        logger.error(
            f"The dataset {dataset.path.name} has no annotation. You can run `cenfind predict` instead"
        )
        sys.exit(2)

    if type(args.tolerance) == int:
        tolerances = [args.tolerance]
    else:
        tolerances = args.tolerance

    path_visualisation_model = dataset.visualisation / args.model.name
    path_visualisation_model.mkdir(exist_ok=True)

    from cenfind.core.detectors import extract_foci, extract_nuclei

    perfs = []
    with open(dataset.path / "test.txt", "r") as f:
        pairs = [l.strip("\n").split(",") for l in f.readlines()]

    for field_name, channel in pairs:
        channel = int(channel)
        field = Field(dataset.projections / f"{field_name}.tif")
        annotation = load_foci(dataset.annotations / "centrioles" / f"{field.name}_C{channel}.txt")
        predictions = extract_foci(field, channel, args.model, prob_threshold=args.threshold)
        nuclei = extract_nuclei(field, args.channel_nuclei)

        for tol in tolerances:
            logger.info("Processing %s %s %s" % (field, channel, tol))
            perf = evaluate(field, channel, annotation, predictions, tol, threshold=args.threshold)
            background = create_vignette(field, marker_index=channel, nuclei_index=args.channel_nuclei)
            vis = visualisation(background=background, centrioles=predictions, nuclei=nuclei)
            tf.imwrite(path_visualisation_model / f"{field.name}_C{channel}_pred.png", vis)
            perfs.append(perf)

    performance_df = pd.DataFrame(perfs)
    performance_df = performance_df.set_index("field")

    if args.performances_file:
        performance_df.to_csv(args.performances_file)
        print(performance_df)
        logger.info("Performances have been saved under %s" % args.performances_file)
    else:
        print(performance_df)
        logger.warning("Performances are ONLY displayed, not saved")


if __name__ == "__main__":
    args = argparse.Namespace(dataset=ROOT_DIR/ 'data/dataset_test',
                              model=ROOT_DIR / 'models/master',
                              channel_nuclei=0,
                              tolerance=3,
                              threshold=.5,
                              performances_file=ROOT_DIR / "out/perfs.txt"
                              )
    run(args)



from typing import List
import logging

import numpy as np
import pandas as pd

from spotipy.utils import points_matching
from types import SimpleNamespace

from cenfind.core.data import Field
from cenfind.core.structures import Centriole, Nucleus

logger = logging.getLogger(__name__)


def evaluate(
        field: Field,
        channel: int,
        annotation: np.ndarray,
        predictions: List[Centriole],
        tolerance: int,
        threshold: float,
) -> dict:
    """
    Compute the accuracy of the prediction on one field.
    :param field:
    :param channel:
    :param annotation:
    :param predictions:
    :param tolerance:
    :param threshold:
    :return: dictionary of metrics for the field
    """
    _predictions = [f.centre_xy for f in predictions]
    if all((len(_predictions), len(annotation))) > 0:
        res = points_matching(annotation, _predictions, cutoff_distance=tolerance)
    else:
        logger.warning(
            "threshold: %f; detected: %d; annotated: %d... Set precision and accuracy to zero"
            % (threshold, len(_predictions), len(annotation))
        )
        res = SimpleNamespace()
        res.precision = 0.0
        res.recall = 0.0
        res.f1 = 0.0
        res.tp = (0,)
        res.fp = (0,)
        res.fn = 0
    perf = {
        "field": field.name,
        "channel": channel,
        "n_actual": len(annotation),
        "n_preds": len(_predictions),
        "threshold": threshold,
        "tolerance": tolerance,
        "tp": res.tp[0] if type(res.tp) == tuple else res.tp,
        "fp": res.fp[0] if type(res.fp) == tuple else res.fp,
        "fn": res.fn[0] if type(res.fn) == tuple else res.fn,
        "precision": np.round(res.precision, 3),
        "recall": np.round(res.recall, 3),
        "f1": np.round(res.f1, 3),
    }
    return perf

def frequency(df: pd.DataFrame) -> pd.DataFrame:
    """
    Counts the absolute frequency of number of centriole per field.

    The bins are 0, 1, 2, 3, 4, +.

    Args:
        df: Dataframe with the number of centriole per nuclei

    Returns: Dataframe with absolute frequencies

    """
    cuts = [0, 1, 2, 3, 4, 5, np.inf]
    labels = "0 1 2 3 4 +".split(" ")

    df = df.loc[df["full_in_field"]]

    result = pd.cut(df["score"], cuts, right=False, labels=labels,
                    include_lowest=True)
    result = result.groupby(["field", "channel"]).value_counts()
    result.name = "freq_abs"
    result = result.reset_index()
    result = result.rename({"score": "score_category"}, axis=1)
    result = result.pivot(index=["channel", "field"], columns="score_category")
    result.columns = labels

    return result


def proportion_cilia(field: Field, cilia: List[Centriole], nuclei: List[Nucleus], channel_cilia: int) -> pd.DataFrame:
    """
    Computes the proportion of ciliated cells in one field of view.

    Args:
        field: The field of view analysed.
        cilia: List of detected cilia.
        nuclei: List of detected nuclei.
        channel_cilia: Channel used for cilia detection.

    Returns: Dataframe with field | channel cilia | n_nuclei | n_cilia | p_ciliated columns.

    """
    proportions = []
    ciliated = len(cilia) / len(nuclei)
    proportions.append({"field": field.name,
                        "channel_cilia": channel_cilia,
                        "n_nuclei": len(nuclei),
                        "n_cilia": len(cilia),
                        "p_ciliated": round(ciliated, 2)})
    return pd.DataFrame(proportions)



import json
import logging
from pathlib import Path
from typing import List

import numpy as np
import pandas as pd
import tifffile as tif
from cenfind.core.structures import Centriole, Nucleus

logger = logging.getLogger(__name__)


def save_assigned(dst: Path, assigned: np.ndarray) -> None:
    """
    Saves the assignment matrix as a text file.

    Args:
        dst: The destination path
        assigned: the assignment matrix

    Returns:
    """

    logger.info("Writing assigned matrix to %s" % str(dst))
    np.savetxt(str(dst), assigned, fmt="%i")


def save_assigned_centrioles(dst: Path, assigned_centrioles) -> None:
    """
    Saves the assigned centrioles as a text file.

    The file is a mapping between the centriole and nucleus indices.

    Args:
        dst: The destination path
        assigned_centrioles: the assignment matrix

    Returns:
    """

    result = pd.DataFrame(assigned_centrioles)
    result.columns = ["centriole_index", "nucleus_index"]
    result.to_csv(dst, sep='\t', index=False)


def save_points(dst: Path, centrioles: List[Centriole]) -> None:
    """Saves the centrioles as a text file.

        The file contains the position of the centrioles in row-major format and signal intensity.

        Args:
           dst: The destination path.
           centrioles: the list of centrioles.

    Returns:
    """
    if len(centrioles) == 0:
        result = pd.DataFrame([])
        logger.info("No centriole detected")
    else:
        result = pd.DataFrame.from_dict({c.index: c.as_dict() for c in centrioles}, orient='index')
    result.to_csv(dst, index_label="index", index=False, sep='\t')


def save_contours(dst: Path, nuclei: List[Nucleus]) -> None:
    """
       Saves the centrioles as a text file.

       The file contains the position of the centrioles in row-major format and signal intensity.

       Args:
           dst: The destination path.
           nuclei: the list of centrioles.

       Returns:
    """
    with open(dst, "w") as file:
        json.dump({"nuclei": {nucleus.index: nucleus.as_dict() for nucleus in nuclei}}, file)
        logger.info("Writing contours to %s" % str(dst))


def save_visualisation(dst: Path, vis: np.ndarray) -> None:
    """
    Saves the visualisation image.

    Args:
       dst: The destination path.
       vis: The image.

    """
    logger.info("Writing visualisation to %s" % (str(dst)))
    tif.imwrite(dst, vis)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['cenfind',
 'cenfind.cli',
 'cenfind.core',
 'cenfind.publication',
 'cenfind.training']

package_data = \
{'': ['*']}

install_requires = \
['albumentations>=1.3.0,<2.0.0',
 'attrs>=22.2.0,<23.0.0',
 'csbdeep>=0.7.3,<0.8.0',
 'llvmlite==0.39.1',
 'numba==0.56.4',
 'numpy>=1.23.5,<2.0.0',
 'opencv-python>=4.7.0.72,<5.0.0.0',
 'ortools==9.4.1874',
 'pandas>=1.4.1,<2.0.0',
 'protobuf==3.19.6',
 'scikit-image>=0.19.2,<0.20.0',
 'scikit-learn>=1.2.1,<2.0.0',
 'scipy>=1.7.1,<2.0.0',
 'seaborn>=0.13.0,<0.14.0',
 'spotipy-detector>=0.1.0,<0.2.0',
 'stardist>=0.8.3,<0.9.0',
 'tifffile>=2022.5.4,<2023.0.0',
 'tqdm>=4.62.3,<5.0.0']

extras_require = \
{':sys_platform == "darwin"': ['tensorflow-macos==2.9.0',
                               'tensorflow-metal==0.5.0'],
 ':sys_platform == "win32" or sys_platform == "linux"': ['tensorflow==2.9.0']}

entry_points = \
{'console_scripts': ['cenfind = cenfind.__main__:main']}

setup_kwargs = {
    'name': 'cenfind',
    'version': '0.15.2',
    'description': 'Score cells for centrioles in IF data',
    'long_description': '![alt text](figures/logos/cenfind_logo_full_dark.png)\n\n# CenFind\n\n**Cenfind** is a command line interface written in Python to detect and assign centrioles in immunofluorescence images of human cells.\nSpecifically, it orchestrates the detection of centrioles, the detection of the nuclei and the assignment of the centrioles to the nearest nucleus.\n\n## Getting started\n\n1. Install cenfind from PyPI:\n```shell\npip install cenfind\n```\n2. You need to download it from https://figshare.com/articles/software/Cenfind_model_weights/21724421\n3. Collect all images in a project folder inside a projections folder (<project_name>/projection/).\n4. Run `score` with the path to the project, the path to the model, the index of the nuclei channel (usually 0 or 3),\n   the channel to score:\n\n```shell\ncenfind score /path/to/dataset /path/to/model/ -n 0 -c 1 2 3\n```\n\n5. Check that the predictions in the folders `visualisations/` and `statistics/`\n\nFor more information, please check the documentation (https://cenfind.readthedocs.io).\n\n## Citation\n\nWe appreciate citations as they help us obtain grant funding and let us discover its application range.\n\nTo cite Cenfind in publications, please use:\n\nBürgy, L., Weigert, M., Hatzopoulos, G. et al. CenFind: a deep-learning pipeline for efficient centriole detection in\nmicroscopy datasets. BMC Bioinformatics 24, 120 (2023). https://doi.org/10.1186/s12859-023-05214-2\n',
    'author': 'Leo Burgy',
    'author_email': 'leo.burgy@epfl.ch',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/UPGON/cenfind',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)
